/* Copyright (C) 2008 and 2009 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the src/utils sub-directory);
   if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file should by reason
only of that instantiation be subject to the restrictions of use in
the GNU Lesser General Public License.  With that in mind, the words
"and instantiations of templates (of any length)" shall be treated as
inserted in the fourth paragraph of section 5 of that licence after
the words "and small inline functions (ten lines or less in length)".
This does not affect any other reason why object code may be subject
to the restrictions in that licence (nor for the avoidance of doubt
does it affect the application of section 2 of that licence to
modifications of the source code in this file).

*/

#ifndef CALLBACK_H
#define CALLBACK_H

/* These callback classes encapsulate callbacks (they are closures).
   They comprise a generic callback creation and execution interface.
   There is a basic Callback::Callback type, which is a pure closure
   where all the arguments are bound in the constructor and is
   completely opaque.  CallbackArg<T> is a class which takes one
   argument of type T when the callback is dispatched, with any other
   argument being bound at construction time - the type parameter is
   the type of the unbound argument.  (The pure Callback::Callback
   type is in fact just a typedef for Callback::CallbackArg<void>: the
   two types are interchangeable.)  The function or method referred to
   must be one of void return type.

   They are particularly useful where a callback object may need to be
   handed between threads and where using sigc::slot would therefore
   be unsafe - this is because sigc::slot objects representing
   non-static methods of a class derived from sigc::trackable may
   invoke sigc::trackable functions at slot construction and
   destruction which are not thread safe.

   The pure Callback::Callback class can also be useful as the data
   argument of a call to g_signal_connect_data(), in a case where a
   better design arises when passing arguments known at connect time
   by storing them in the callback object itself (say, where otherwise
   you would pass a plain struct as the data argument), or for general
   event passing when used together with g_idle_add_full() from glib
   (see the Callback::post() functions below).

   These classes are also used in the Emitter/EmitterArg classes in
   emitter.h, which enable callbacks to be connected to an emitter and
   provide for automatic disconnection where a class object whose
   member a callback represents ceases to exist.

   The templated helper Callback::make() functions make it trivial to
   create a callback object of the correct type.  A maximum of two
   bound arguments to pass to the relevant function or class method is
   provided for and a maximum of one free argument to pass at call
   time - but this can be extended indefinitely.  If there is a free
   argument, it must be the last argument of the relevant function or
   method to be called.  Note that bound arguments of a relevent class
   method should not comprise a reference argument (T& or const T&).
   Although the types of the template parameters of the
   Callback::make() call can be qualified by hand to avoid a type
   mismatch with a reference, it will usually be unsafe to do so.

   If a functor class is required (say for passing to a c++ algorithm
   or container, or for automatic lifetime management of the Callback
   object), the Callback::Functor and Callback::FunctorArg wrapper
   classes can be used.  Callback::SafeFunctor
   Callback::SafeFunctorArg classes are the same, but may be passed
   between threads.  However, for many Std C++ algorithms where
   anonymous functors created as temporaries can be used, the
   std::ptr_fun() and MemFun::make() factory functions will be a more
   obvious choice (for the latter, see the mem_func.h file).

   Usage:

   For a class my_obj of type MyObj, with a method void
   MyObj::my_method(int, const char*), usage for a fully bound
   callback or functor would be:
   
     int arg = 1;
     Callback::Callback* cb =
       Callback::make(my_obj, &MyObj::my_method, arg, "Hello\n");
     cb->dispatch();
     delete cb;

     Callback::Functor f(Callback::make(my_obj, &MyObj::my_method, arg, "Hello\n"));
     f();

   Or for a partially bound callback or functor:

     Callback::CallbackArg<const char*>* cb =
       Callback::make(my_obj, &MyObj::my_method, arg);
     cb->dispatch("Hello\n");
     delete cb;

     Callback::FunctorArg<const char*> f(Callback::make(my_obj, &MyObj::my_method, arg));
     f("Hello\n");

   Callback/CallbackArg classes do not provide for a return value.  If
   a result is wanted, users should pass an argument by pointer (or
   pointer to pointer).
*/

#include "shared_ptr.h"

#include <functional> // for std::less

#include <glib.h>

namespace Callback {

/* The following Callback and CallbackArg types comprise the interface
   classes that users will generally see, and are suitable for passing
   by void*, which if using glib and/or gthreads and/or pthreads is
   almost certain to happen at some stage.

   Because of the classes' intended usage, the dispatch() method
   returns void.  The CallbackArg class could be additionally
   templated to provide a return value, but that would affect the
   simplicity of the interface, and if a case were to arise where a
   result is needed, an alternative is for users to pass an argument
   by pointer (or pointer to pointer) rather than have a return value.
*/

/* Declare the two basic interface types */

template <class FreeArg> class CallbackArg;
typedef CallbackArg<void> Callback;

/* now the class definitions */

template <class FreeArg>
class CallbackArg {
public:
  // because we will usually be calling through a base class pointer,
  // it would be more natural to invoke the call back through a normal
  // member function rather than by operator()() - use the FunctorArg
  // wrapper class for the latter
  virtual void dispatch(FreeArg) const = 0;
  virtual ~CallbackArg() {};
};

template <>
class CallbackArg<void> {
public:
  virtual void dispatch() const = 0;
  virtual ~CallbackArg() {};
};

/* The four basic functor types */

template <class FreeArg> class FunctorArg;
template <class FreeArg> class SafeFunctorArg;
typedef FunctorArg<void> Functor;
typedef SafeFunctorArg<void> SafeFunctor;

/* Functor friend functions */

// we can use built-in operator == when comparing pointers referencing
// different objects of the same type
template <class T>
bool operator==(const FunctorArg<T>& f1, const FunctorArg<T>& f2) {
  return (f1.cb_s.get() == f2.cb_s.get());
}

// we must use std::less rather than the < built-in operator for
// pointers to objects not within the same array or object: "For
// templates greater, less, greater_equal, and less_equal, the
// specializations for any pointer type yield a total order, even if
// the built-in operators <, >, <=, >= do not." (para 20.3.3/8).
template <class T>
bool operator<(const FunctorArg<T>& f1, const FunctorArg<T>& f2) {
  return std::less<CallbackArg<T>*>(f1.cb_s.get(), f2.cb_s.get());
}

template <class T>
bool operator==(const SafeFunctorArg<T>& f1, const SafeFunctorArg<T>& f2) {
  return (f1.cb_s.get() == f2.cb_s.get());
}

template <class T>
bool operator<(const SafeFunctorArg<T>& f1, const SafeFunctorArg<T>& f2) {
  return std::less<CallbackArg<T>*>(f1.cb_s.get(), f2.cb_s.get());
}

/* the functor classes */

// in the Functor wrapper class, the callback object is kept by
// SharedPtr so the functor can be copied and offers automatic
// lifetime management of the wrapped callback object.  Ownership is
// taken of the CallbackArg object passed to the constructor taking a
// CallbackArg pointer, so that constructor should be treated like a
// shared pointer constructor - only pass a newly allocated object to
// it (or copy construct it or assign to it from another existing
// FunctorArg object.).
template <class FreeArg>
class FunctorArg {
  SharedPtr<const CallbackArg<FreeArg> > cb_s;
public:
  void operator()(FreeArg arg) const {if (cb_s.get()) cb_s->dispatch(arg);}

  friend bool operator== <>(const FunctorArg&, const FunctorArg&);
  friend bool operator< <>(const FunctorArg&, const FunctorArg&);

  // don't make this constructor explicit - we can then assign a
  // CallbackArg<FreeArg> object directly to the functor object
  FunctorArg(const CallbackArg<FreeArg>* cb_p): cb_s(cb_p) {}
  FunctorArg(const FunctorArg& f): cb_s(f.cb_s) {}
  FunctorArg() {};
};

template <>
class FunctorArg<void> {
  SharedPtr<const Callback> cb_s;
public:
  void operator()() const {if (cb_s.get()) cb_s->dispatch();}

  friend bool operator== <>(const FunctorArg&, const FunctorArg&);
  friend bool operator< <>(const FunctorArg&, const FunctorArg&);

  FunctorArg(const Callback* cb_p): cb_s(cb_p) {}
  FunctorArg(const FunctorArg& f): cb_s(f.cb_s) {}
  FunctorArg() {};
};

// this class is the same as Callback::FunctorArg except that it will
// provide synchronisation of the reference count between threads.
// Use it where a functor wrapper object is to be passed between
// threads.
template <class FreeArg>
class SafeFunctorArg {
  SharedLockPtr<const CallbackArg<FreeArg> > cb_s;
public:
  void operator()(FreeArg arg) const {if (cb_s.get()) cb_s->dispatch(arg);}

  friend bool operator== <>(const SafeFunctorArg&, const SafeFunctorArg&);
  friend bool operator< <>(const SafeFunctorArg&, const SafeFunctorArg&);

  SafeFunctorArg(const CallbackArg<FreeArg>* cb_p): cb_s(cb_p) {}
  SafeFunctorArg(const SafeFunctorArg& f): cb_s(f.cb_s) {}
  SafeFunctorArg() {};
};

template <>
class SafeFunctorArg<void> {
  SharedLockPtr<const Callback> cb_s;
public:
  void operator()() const {if (cb_s.get()) cb_s->dispatch();}

  friend bool operator== <>(const SafeFunctorArg&, const SafeFunctorArg&);
  friend bool operator< <>(const SafeFunctorArg&, const SafeFunctorArg&);

  SafeFunctorArg(const Callback* cb_p): cb_s(cb_p) {}
  SafeFunctorArg(const SafeFunctorArg& f): cb_s(f.cb_s) {}
  SafeFunctorArg() {};
};

/* the callback implementation classes */

template <class T>
class Callback0: public Callback {
public:
  typedef void (T::* MemFunc)();
private:
  T* obj;
  MemFunc func;
public:
  void dispatch() const {(obj->*func)();}
  Callback0(T& obj_, MemFunc func_): obj(&obj_), func(func_) {}
};

template <class T, class FreeArg>
class CallbackArg0: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(FreeArg);
private:
  T* obj;
  MemFunc func;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(free_arg);}
  CallbackArg0(T& obj_, MemFunc func_): obj(&obj_), func(func_) {}
};

template <class T, class BoundArg1>
class Callback1: public Callback {
public:
  typedef void (T::* MemFunc)(BoundArg1);
private:
  T* obj;
  MemFunc func;
  BoundArg1 arg1;
public:
  void dispatch() const {(obj->*func)(arg1);}
  Callback1(T& obj_, MemFunc func_, BoundArg1 arg1_): obj(&obj_), func(func_), arg1(arg1_) {}
};

template <class T, class BoundArg1, class FreeArg>
class CallbackArg1: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(BoundArg1, FreeArg);
private:
  T* obj;
  MemFunc func;
  BoundArg1 arg1;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(arg1, free_arg);}
  CallbackArg1(T& obj_, MemFunc func_, BoundArg1 arg1_): obj(&obj_), func(func_), arg1(arg1_) {}
};

template <class T, class BoundArg1, class BoundArg2>
class Callback2: public Callback {
public:
  typedef void (T::* MemFunc)(BoundArg1, BoundArg2);
private:
  T* obj;
  MemFunc func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch() const {(obj->*func)(arg1, arg2);}
  Callback2(T& obj_, MemFunc func_, BoundArg1 arg1_, BoundArg2 arg2_): obj(&obj_), func(func_),
                                                                       arg1(arg1_), arg2(arg2_) {}
};

template <class T, class BoundArg1, class BoundArg2, class FreeArg>
class CallbackArg2: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(BoundArg1, BoundArg2, FreeArg);
private:
  T* obj;
  MemFunc func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(arg1, arg2, free_arg);}
  CallbackArg2(T& obj_, MemFunc func_, BoundArg1 arg1_, BoundArg2 arg2_): obj(&obj_), func(func_),
                                                                          arg1(arg1_), arg2(arg2_) {}
};

/* const versions, for binding to const methods */

template <class T>
class Callback0_const: public Callback {
public:
  typedef void (T::* MemFunc)() const;
private:
  const T* obj;
  MemFunc func;
public:
  void dispatch() const {(obj->*func)();}
  Callback0_const(const T& obj_, MemFunc func_): obj(&obj_), func(func_) {}
};

template <class T, class FreeArg>
class CallbackArg0_const: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(FreeArg) const;
private:
  const T* obj;
  MemFunc func;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(free_arg);}
  CallbackArg0_const(const T& obj_, MemFunc func_): obj(&obj_), func(func_) {}
};

template <class T, class BoundArg1>
class Callback1_const: public Callback {
public:
  typedef void (T::* MemFunc)(BoundArg1) const;
private:
  const T* obj;
  MemFunc func;
  BoundArg1 arg1;
public:
  void dispatch() const {(obj->*func)(arg1);}
  Callback1_const(const T& obj_, MemFunc func_, BoundArg1 arg1_): obj(&obj_), func(func_), arg1(arg1_) {}
};

template <class T, class BoundArg1, class FreeArg>
class CallbackArg1_const: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(BoundArg1, FreeArg) const;
private:
  const T* obj;
  MemFunc func;
  BoundArg1 arg1;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(arg1, free_arg);}
  CallbackArg1_const(const T& obj_, MemFunc func_, BoundArg1 arg1_): obj(&obj_), func(func_), arg1(arg1_) {}
};

template <class T, class BoundArg1, class BoundArg2>
class Callback2_const: public Callback {
public:
  typedef void (T::* MemFunc)(BoundArg1, BoundArg2) const;
private:
  const T* obj;
  MemFunc func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch() const {(obj->*func)(arg1, arg2);}
  Callback2_const(const T& obj_, MemFunc func_, BoundArg1 arg1_, BoundArg2 arg2_): obj(&obj_), func(func_),
                                                                                   arg1(arg1_), arg2(arg2_) {}
};

template <class T, class BoundArg1, class BoundArg2, class FreeArg>
class CallbackArg2_const: public CallbackArg<FreeArg> {
public:
  typedef void (T::* MemFunc)(BoundArg1, BoundArg2, FreeArg) const;
private:
  const T* obj;
  MemFunc func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch(FreeArg free_arg) const {(obj->*func)(arg1, arg2, free_arg);}
  CallbackArg2_const(const T& obj_, MemFunc func_, BoundArg1 arg1_, BoundArg2 arg2_): obj(&obj_), func(func_),
                                                                                      arg1(arg1_), arg2(arg2_) {}
};

/* for static class methods and non-class functions */

class Callback0_static: public Callback {
public:
  typedef void (*Func)();
private:
  Func func;
public:
  void dispatch() const {func();}
  Callback0_static(Func func_): func(func_) {}
};

template <class FreeArg>
class CallbackArg0_static: public CallbackArg<FreeArg> {
public:
  typedef void (*Func)(FreeArg);
private:
  Func func;
public:
  void dispatch(FreeArg free_arg) const {func(free_arg);}
  CallbackArg0_static(Func func_): func(func_) {}
};

template <class BoundArg1>
class Callback1_static: public Callback {
public:
  typedef void (*Func)(BoundArg1);
private:
  Func func;
  BoundArg1 arg1;
public:
  void dispatch() const {func(arg1);}
  Callback1_static(Func func_, BoundArg1 arg1_): func(func_), arg1(arg1_) {}
};

 template <class BoundArg1, class FreeArg>
class CallbackArg1_static: public CallbackArg<FreeArg> {
public:
  typedef void (*Func)(BoundArg1, FreeArg);
private:
  Func func;
  BoundArg1 arg1;
public:
  void dispatch(FreeArg free_arg) const {func(arg1, free_arg);}
  CallbackArg1_static(Func func_, BoundArg1 arg1_): func(func_), arg1(arg1_) {}
};

template <class BoundArg1, class BoundArg2>
class Callback2_static: public Callback {
public:
  typedef void (*Func)(BoundArg1, BoundArg2);
private:
  Func func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch() const {func(arg1, arg2);}
  Callback2_static(Func func_, BoundArg1 arg1_, BoundArg2 arg2_): func(func_),
			                                          arg1(arg1_), arg2(arg2_) {}
};

 template <class BoundArg1, class BoundArg2, class FreeArg>
class CallbackArg2_static: public CallbackArg<FreeArg> {
public:
  typedef void (*Func)(BoundArg1, BoundArg2, FreeArg);
private:
  Func func;
  BoundArg1 arg1;
  BoundArg2 arg2;
public:
  void dispatch(FreeArg free_arg) const {func(arg1, arg2, free_arg);}
  CallbackArg2_static(Func func_, BoundArg1 arg1_, BoundArg2 arg2_): func(func_),
			                                             arg1(arg1_), arg2(arg2_) {}
};


/* Convenience functions making callback objects on freestore.  These
 * can for example be passed as the first argument of the
 * Thread::start() method in thread.h. They are also used by the
 * Callback::post() function.
*/

template <class T>
Callback* make(T& t,
	       void (T::*func)()) {
  return new Callback0<T>(t, func);
}

template <class T, class FreeArg>
CallbackArg<FreeArg>* make(T& t,
			   void (T::*func)(FreeArg)) {
  return new CallbackArg0<T, FreeArg>(t, func);
}

template <class T, class BoundArg1>
Callback* make(T& t,
	       void (T::*func)(BoundArg1),
	       BoundArg1 arg1) {
  return new Callback1<T, BoundArg1>(t, func, arg1);
}

template <class T, class BoundArg1, class FreeArg>
CallbackArg<FreeArg>* make(T& t,
			   void (T::*func)(BoundArg1, FreeArg),
			   BoundArg1 arg1) {
  return new CallbackArg1<T, BoundArg1, FreeArg>(t, func, arg1);
}

template <class T, class BoundArg1, class BoundArg2>
Callback* make(T& t,
	       void (T::*func)(BoundArg1, BoundArg2),
	       BoundArg1 arg1,
	       BoundArg2 arg2) {
  return new Callback2<T, BoundArg1, BoundArg2>(t, func, arg1, arg2);
}

template <class T, class BoundArg1, class BoundArg2, class FreeArg>
CallbackArg<FreeArg>* make(T& t,
			   void (T::*func)(BoundArg1, BoundArg2, FreeArg),
			   BoundArg1 arg1,
			   BoundArg2 arg2) {
  return new CallbackArg2<T, BoundArg1, BoundArg2, FreeArg>(t, func, arg1, arg2);
}

/* const versions, for binding to const methods */

template <class T>
Callback* make(const T& t,
	       void (T::*func)() const) {
  return new Callback0_const<T>(t, func);
}

template <class T, class FreeArg>
CallbackArg<FreeArg>* make(const T& t,
			   void (T::*func)(FreeArg) const) {
  return new CallbackArg0_const<T, FreeArg>(t, func);
}

template <class T, class BoundArg1>
Callback* make(const T& t,
	       void (T::*func)(BoundArg1) const,
	       BoundArg1 arg1) {
  return new Callback1_const<T, BoundArg1>(t, func, arg1);
}

template <class T, class BoundArg1, class FreeArg>
CallbackArg<FreeArg>* make(const T& t,
			   void (T::*func)(BoundArg1, FreeArg) const,
			   BoundArg1 arg1) {
  return new CallbackArg1_const<T, BoundArg1, FreeArg>(t, func, arg1);
}

template <class T, class BoundArg1, class BoundArg2>
Callback* make(const T& t,
	       void (T::*func)(BoundArg1, BoundArg2) const,
	       BoundArg1 arg1,
	       BoundArg2 arg2) {
  return new Callback2_const<T, BoundArg1, BoundArg2>(t, func, arg1, arg2);
}

template <class T, class BoundArg1, class BoundArg2, class FreeArg>
CallbackArg<FreeArg>* make(const T& t,
			   void (T::*func)(BoundArg1, BoundArg2, FreeArg) const,
			   BoundArg1 arg1,
			   BoundArg2 arg2) {
  return new CallbackArg2_const<T, BoundArg1, BoundArg2, FreeArg>(t, func, arg1, arg2);
}

/* for static class methods and non-class functions */

inline Callback* make(void (*func)()) {
  return new Callback0_static(func);
}

template <class FreeArg>
CallbackArg<FreeArg>* make(void (*func)(FreeArg)) {
  return new CallbackArg0_static<FreeArg>(func);
}

template <class BoundArg1>
Callback* make(void (*func)(BoundArg1),
	       BoundArg1 arg1) {
  return new Callback1_static<BoundArg1>(func, arg1);
}

template <class BoundArg1, class FreeArg>
CallbackArg<FreeArg>* make(void (*func)(BoundArg1, FreeArg),
			   BoundArg1 arg1) {
  return new CallbackArg1_static<BoundArg1, FreeArg>(func, arg1);
}

template <class BoundArg1, class BoundArg2>
Callback* make(void (*func)(BoundArg1, BoundArg2),
	       BoundArg1 arg1,
	       BoundArg2 arg2) {
  return new Callback2_static<BoundArg1, BoundArg2>(func, arg1, arg2);
}

template <class BoundArg1, class BoundArg2, class FreeArg>
CallbackArg<FreeArg>* make(void (*func)(BoundArg1, BoundArg2, FreeArg),
			   BoundArg1 arg1,
			   BoundArg2 arg2) {
  return new CallbackArg2_static<BoundArg1, BoundArg2, FreeArg>(func, arg1, arg2);
}


/*
The Callback::post() function will execute a callback in the glib main
loop and can be used (amongst other things) to pass an event from a
worker thread to the main program thread.  In that respect, it
provides an alternative to the Notifier class in
notifier.h/notifier.cpp.  It is passed a pointer to a
Callback::Callback object created with a call to Callback::make.

To provide for thread-safe automatic disconnection of the callback if
the object whose method it represents is destroyed before the callback
executes in the main loop, include a Releaser as a public member of
that object and pass the Releaser object as the second argument of
Callback::post().  See file emitter.h for further details.  Note that
for this to be race free, the lifetime of the remote object whose
method is to be invoked must be determined by the main loop thread -
when the main loop begins invoking the execution of the callback, the
remote object must either wholly exist (in which case the callback
will be invoked) or have been destroyed (in which case the callback
will be ignored), and not be in some transient half-state governed by
another thread.

Advantages as against Notifier:

1. If there are a lot of different events requiring callbacks to be
   dispatched in the program from worker threads to the main thread,
   this avoids having separate Notifier objects for each event.

2. It is easier to pass arguments with varying values - they can be
   passed as templated arguments to the Callback::make or
   Callback::make_static functions and no special synchronisation is
   normally required (the call to g_idle_add() invokes locking of the
   main loop which will have the effect of ensuring memory
   visibility).  With a Notifier object it may be necessary to use an
   asynchronous queue to pass variable values (or to bind a reference
   to the data, thus normally requiring separate synchronisation).

Disadvantages as against Notifier:

1. Less efficient, as a new callback object has to be created on
   freestore every time the callback is invoked, together with a new
   Emitter object if a Releaser is used to track the callback.

2. Multiple callbacks relevant to a single event cannot be invoked
   from a single call for the event - each callback has to be
   separately dispatched.
*/


class Releaser;

/* in ascending order of priorities, priorities are G_PRIORITY_LOW,
   G_PRIORITY_DEFAULT_IDLE, G_PRIORITY_HIGH_IDLE, G_PRIORITY_DEFAULT
   and G_PRIORITY_HIGH. The default is G_PRIORITY_DEFAULT_IDLE.  This
   determines the order in which the callback will appear in the event
   list in the main loop, not the priority which the OS will adopt
*/
void post(const Callback* cb, gint priority = G_PRIORITY_DEFAULT_IDLE);

void post(const Callback* cb, Releaser&, gint priority = G_PRIORITY_DEFAULT_IDLE);

} // namespace Callback

#endif
