/* Copyright (C) 2009 Chris Vine

The library comprised in this file or of which this file is part is
distributed by Chris Vine under the GNU Lesser General Public
License as follows:

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public License
   as published by the Free Software Foundation; either version 2.1 of
   the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License, version 2.1, for more details.

   You should have received a copy of the GNU Lesser General Public
   License, version 2.1, along with this library (see the file LGPL.TXT
   which came with this source code package in the src/utils sub-directory);
   if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA.

However, it is not intended that the object code of a program whose
source code instantiates a template from this file should by reason
only of that instantiation be subject to the restrictions of use in
the GNU Lesser General Public License.  With that in mind, the words
"and instantiations of templates (of any length)" shall be treated as
inserted in the fourth paragraph of section 5 of that licence after
the words "and small inline functions (ten lines or less in length)".
This does not affect any other reason why object code may be subject
to the restrictions in that licence (nor for the avoidance of doubt
does it affect the application of section 2 of that licence to
modifications of the source code in this file).

*/

#ifndef EMITTER_H
#define EMITTER_H

/* A Callback::Emitter or Callback::EmitterArg object is a list of
   Callback::Functor or Callback::FunctorArg objects.
   Functor/FunctorArg objects may be "connected" to the
   Emitter/EmitterArg object, and all functors so connected will be
   executed when the operator()() or emit() member functions of the
   Emitter/EmitterArg object concerned is called.  They will be
   called in the order in which they were connected.

   Emitter is a typedef for EmitterArg<void>.  The generalised
   EmitterArg<T> type contains FunctorArg<T> objects (type T being the
   unbound argument of a CallbackArg<T> callback - see callback.h for
   further details.)  The Emitter type holds Functor (namely
   FunctorArg<void>) objects.

   The main advantage of an emitter object as opposed to storing a
   functor object directly, apart from the fact that more than one
   functor can be dispatched by a single call to EmitterArg<>::emit()
   or EmitterArg<>::operator()(), is that it provides for automatic
   disconnection of a functor if the object whose member function it
   represents has ceased to exist.

   Where automatic disconnection is wanted, the object whose method is
   to be encapsulated by a functor must have a Releaser object as a
   public member function.  The Releaser object should be passed as
   the second argument of EmitterArg<>::connect().  As well as a
   Releaser object informing an emitter object when it has ceased to
   exist, an emitter object will do the same to the Releaser object if
   the emitter object happens to be destroyed before an object whose
   members it references (and therefore before the Releaser object).
   Automatic disconnection is mainly useful for non-static member
   functions, but it can be employed for static member functions if
   wanted (that will in effect bind the lifetime of the functor to
   that of the object to whose Releaser the functor has been
   attached.)

   It is safe for a connected function (i) to delete the EmitterArg<>
   object to which it is connected, even if there are other functors
   still to execute in the same emission (which will execute normally
   provided they do not try to call any of the emitter's functions),
   (ii) to call 'delete this' nothwithstanding that the connected
   function is protected by a Releaser object (assuming all the other
   restraints on calling 'delete this' are met), provided that no
   other access would be made to the deleted object in a function call
   connected to the same emitter which is due to execute subsequently
   in the same emission, and (iii) to disconnect itself from the
   EmitterArg<> object.  This design approach has a trade-off: if a
   connected function tries to block, unblock or disconnect another
   function connected to the same EmitterArg<> object which is due to
   execute subsequently in the same emission (or to block, unblock or
   disconnect itself when it is due to execute again subsequently in
   the same emission), the attempted block, unblock or disconnection
   will not have any effect on that emission (it will only have effect
   on a subsequent emission).  In addition, a connected function may
   not destroy an object whose non-static method is connected to the
   same emitter and which would execute subsequently in the same
   emission, even if that object is protected by a Releaser object
   (the non-static method will unsuccessfully attempt to execute
   notwithstanding the destruction of the object it would be operating
   on).

   The SafeEmitterArg<> classes are the same as their EmitterArg<>
   counterparts except that they contain SafeFunctorArg<> objects, and
   their emit(), operator()(), connect(), disconnect(), block(),
   unblock() and destructor methods are protected by a mutex so that
   different threads can call these methods on the same emitter
   object, or create and delete the object.

   Note that the mutexes are released when the operator()() methods of
   the relevent SafeFunctorArg<> objects are called, as
   SafeEmitterArg<> objects have no idea what the referenced callbacks
   will do so if they were not released deadlocks could arise from
   recursive or out-of-order locking of the SafeEmitterArg<> mutex.
   It is therefore for users to provide additional synchronisation if
   the functions encapsulated by the relevant functors themselves need
   additional protection.  Note also the subsidiary thread-safety
   points mentioned below.

   The Releaser class is intrinsically thread safe (the overhead of
   locking is so low that it is pointless having a separate
   unprotected class).  This means that if a program is
   multi-threaded, you can use the plain EmitterArg<> classes provided
   that only the thread which creates a particular EmitterArg<> object
   calls connect(), block(), unblock((), emit() or operator()() on it,
   or deletes it, or calls disconnect() on it (either directly or
   through a Releaser object being destroyed).  Where more than one
   thread might do that in relation to any one emitter object, use
   SafeEmitterArg<>.

   These classes are intended as a lightweight thread-safe signal/slot
   mechanism for GUI programming.  For more demanding usage libsigc++
   is a good choice, except that it is not thread-safe.  An
   alternative to libsigc++ is the boost::signal2 module, which is
   thread-safe.

   Subsidiary thread-safety points:

   As mentioned, the SafeEmitterArg<> classes are thread safe, and
   their methods can be called in different threads without ill
   effect.  However, there are some things that cannot be done.  Users
   should observe two points.

   First, it has been mentioned that if a connected function blocks,
   unblocks or disconnects another function connected to the same
   emitter object and due to execute subsequently in the same
   emission, the blocking, unblocking or disconnection will not have
   effect in that emission, and that the user may not delete an object
   whose non-static method is due to execute subsequently in the same
   emission.  The same outcome would result if another thread tries to
   do any of these things while an emission is under way.  Another
   thread should therefore leave alone objects connected to a
   SafeEmitterArg<> object from the time of operator()() or emit()
   beginning to the time of it ending, and not try to interfere.

   Secondly, when a Releaser object is passed as the second argument
   to the connect() method of a SafeEmitterArg<> object, the Releaser
   object must remain in existence until the connect() method returns
   or the emitter may be left in an inconsistent state.

   Usage:

   For a class my_obj of type MyObj, with a method void
   MyObj::my_method(int, const char*), usage for a fully bound
   functor and emitter would be:
   
     int arg = 1;
     Callback::Emitter e;
     e.connect(Callback::make(my_obj, &MyObj::my_method, arg, "Hello\n"));
     e();

     Callback::SafeEmitter se;
     se.connect(Callback::make(my_obj, &MyObj::my_method, arg, "Hello\n"));
     se();

   Or for a partially bound functor and emitter:

     Callback::EmitterArg<const char*> e;
     e.connect(Callback::make(my_obj, &MyObj::my_method, arg));
     e("Hello\n");

     Callback::SafeEmitterArg<const char*> se;
     se.connect(Callback::make(my_obj, &MyObj::my_method, arg));
     se("Hello\n");

   EmitterArg<> classes do not provide for a return value.  If a
   result is wanted, users should pass an argument by pointer (or
   pointer to pointer).

   Exception safety:

   Apart from the emit()/operator()() and connect() methods, nothing
   done to an EmitterArg<> object should cause an exception to be
   thrown.  This is because other methods only iterate through a
   std::list object using std::for_each(), std::find() or by hand, and
   the only things done by std::for_each() or after a std::find() or
   iteration is to remove a functor from the list (copying a functor
   and comparing functors never throw, nor does destroying a functor
   provided the destructors of any bound argument type do not throw).
   Thus, an EmitterArg<>/SafeEmitterArg<> and Releaser object should
   never get into an inconsistent state.

   The connect() method could throw a std::bad_alloc exception, either
   on creating new functors or on pushing the functors onto the list.
   However, were it to do so, the method has strong exception safety
   (assuming merely iterating over a list does not throw, as it should
   not).

   The emit()/operator()() methods could throw std::bad_alloc, and so
   far as that is concerned emission of all the connected functions
   will either all succeed or all fail.  In addition, the connected
   functions referenced by the functors held by the emitter might
   throw when executed.  emit()/operator()() do not attempt to catch
   these exceptions as there is nothing they could do with them.  This
   means that although a throwing connected function will not leave
   the EmitterArg<>/SafeEmitterArg<> object in an inconsistent state,
   any other connected functions due to execute subsequently on that
   same emission will not execute.  If that is important in any
   particular case, the user must incorporate logic in the connected
   functions to cater for an exception causing only part execution, or
   must connect only one function to any one signal and "chain"
   emissions by hand so as to do the right thing.
*/

/*
  Define UTILS_EMITTER_DONT_USE_SCHED_YIELD before the pre-processor
  parses this file if you want to use usleep() instead of POSIX
  sched_yield() to spin nicely on an out-of-order trylock() wait.  To
  see whether sched_yield() exists, the program configuration script
  can eg test for the manifest constant _POSIX_PRIORITY_SCHEDULING in
  unistd.h and call sysconf(), or run an autoconf script for
  sched_yield().  However, all recent systems should provide it since
  it is required by the standard.
*/

/*
  Mutex locking heirarchy:

  Some out-of-order locking must take place because of the
  relationship between the Releaser and SafeEmitterArg<> classes.  The
  mutex of Releaser is given the higher priority.  This means that a
  plain EmitterArg<> object will not take any hit from the fact that
  Releaser is also useable with SafeEmitterArg<> objects.

  One consequence is that to avoid deadlocks, it is the
  SafeEmitterArg<> functions which must yield when a deadlock would
  otherwise arise.  Yielding could occur in
  SafeEmitterArg<>::~SafeEmitterArg() and
  SafeEmitterArg<>::disconnect().
*/


#ifdef UTILS_EMITTER_DONT_USE_SCHED_YIELD
#include <unistd.h>
#else
#include <sched.h>
#endif

#include <list>
#include <algorithm>
#include <functional>

#include "callback.h"
#include "mutex.h"

namespace Callback {

/* The four basic emitter types */

template <class FreeArg> class EmitterArg;
template <class FreeArg> class SafeEmitterArg;
typedef EmitterArg<void> Emitter;
typedef SafeEmitterArg<void> SafeEmitter;

/* the Releaser class for tracking */

class Releaser {

  std::list<SafeFunctor> disconnect_list;
  Thread::Mutex mutex;

  // only an EmitterArg or SafeEmitterArg object can access add(), remove and try_remove()
  void add(const SafeFunctor&);
  void remove(SafeFunctor);
  void try_remove(SafeFunctor, int*);
public:
  template <class T> friend class EmitterArg;
  template <class T> friend class SafeEmitterArg;

  // operator=() and the copy constructor should copy nothing from the
  // assignor, because disconnect_list should be empty in the
  // sssignee, as any class containing us does not acquire as assignee
  // any emitter functors representing any of its methods
  Releaser& operator=(const Releaser&);
  Releaser(const Releaser&) {}
  Releaser() {}
  ~Releaser();
};

inline Releaser::~Releaser(void) {

  Thread::Mutex::Lock lock(mutex);
  // we don't have lambdas in C++03, so use a struct in function scope
  struct DisconnectItem {
    static void exec(const SafeFunctor& f) {f();}
  };
  std::for_each(disconnect_list.begin(), disconnect_list.end(),
		DisconnectItem::exec);

}

inline Releaser& Releaser::operator=(const Releaser&) {
  Thread::Mutex::Lock lock(mutex);
  disconnect_list.clear();
  return *this;
}

inline void Releaser::add(const SafeFunctor& f) {
  Thread::Mutex::Lock lock(mutex);
  disconnect_list.push_back(f);
}

inline void Releaser::remove(SafeFunctor f) {
  Thread::Mutex::Lock lock(mutex);
  std::list<SafeFunctor>::iterator iter =
    std::find(disconnect_list.begin(), disconnect_list.end(), f);

  if (iter != disconnect_list.end()) {
    disconnect_list.erase(iter);
  }
}

// this method is called instead of Releaser::remove() when called by a
// SafeEmitterArg<> object which has locked its mutex, so as to avoid
// out of order locking deadlocks.  It is primarily intended to cover a
// case where both the Releaser and SafeEmitterArg<> destructors are
// operating and will cause the SafeEmitterArg<> destructor to spin
// until the Releaser destructor has done its business, although it is
// also used in SafeEmitterArg<>::disconnect()
inline void Releaser::try_remove(SafeFunctor f, int* result_p) {

  if (!(*result_p = mutex.trylock())) {
    Thread::Mutex::Lock lock(mutex, Thread::locked);
    std::list<SafeFunctor>::iterator iter =
      std::find(disconnect_list.begin(), disconnect_list.end(), f);

    if (iter != disconnect_list.end()) {
      disconnect_list.erase(iter);
    }
  }
}

/* the emitter classes */

template <class FreeArg>
class EmitterArg {

  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed
  // before the remote object is
  struct ListItem {
    FunctorArg<FreeArg> f1;
    Functor f2;
    bool blocked;
    ListItem(FunctorArg<FreeArg> f1_, Functor f2_): f1(f1_), f2(f2_), blocked(false) {}
  };
  
  std::list<ListItem> emission_list;

  // only Releaser objects can access this
  void tracking_disconnect(FunctorArg<FreeArg>);

  // emitters cannot be copied
  EmitterArg(const EmitterArg&);
  EmitterArg& operator=(const EmitterArg&);
public:
  friend class Releaser;

  void operator()(FreeArg arg) const {emit(arg);}
  void emit(FreeArg arg) const;

  // connect() returns the connected functor for convenience
  FunctorArg<FreeArg> connect(const FunctorArg<FreeArg>&);
  FunctorArg<FreeArg> connect(const FunctorArg<FreeArg>&, Releaser&);
  void disconnect(const FunctorArg<FreeArg>&);

  void block(const FunctorArg<FreeArg>&);
  void unblock(const FunctorArg<FreeArg>&);

  EmitterArg() {}
  ~EmitterArg();
};

template <class FreeArg>
EmitterArg<FreeArg>::~EmitterArg() {

  struct DisconnectReleaserItem {
    static void exec(const ListItem& l) {(l.f2)();}
  };
  std::for_each(emission_list.begin(), emission_list.end(),
		DisconnectReleaserItem::exec);
}

template <class FreeArg>
void EmitterArg<FreeArg>::emit(FreeArg arg) const {

  // create a local copy of emission_list, to enable a connected
  // function (i) to delete the EmitterArg<> object to which it is
  // connected, even if there are other functors still to execute in
  // the same emission (which will execute normally provided they do
  // not try to call any of the emitter's functions), (ii) to call
  // 'delete this' nothwithstanding that the connected function is
  // protected by a Releaser object (assuming all the other restraints
  // on calling 'delete this' are met), provided that no other access
  // would be made to the deleted object in a function call connected
  // to the same emitter which is due to execute subsequently in the
  // same emission, and (iii) to disconnect itself from the
  // EmitterArg<> object.  This design approach has a trade-off: if a
  // connected function tries to block, unblock or disconnect another
  // function connected to the same EmitterArg<> object which is due
  // to execute subsequently in the same emission (or to block,
  // unblock or disconnect itself when it is due to execute again
  // subsequently in the same emission), the attempted block, unblock
  // or disconnection will not have any effect on that emission (it
  // will only have effect on a subsequent emission).  In addition, a
  // connected function may not destroy an object whose non-static
  // method is connected to the same emitter and which would execute
  // subsequently in the same emission, even if that object is
  // protected by a Releaser object (the non-static method will
  // unsuccessfully attempt to execute notwithstanding the destruction
  // of the object it would be operating on).

  std::list<ListItem> local_list = emission_list;
  // we don't have lambdas in C++03, so use a struct in function scope
  struct EmitItem {
    static void exec(ListItem l, FreeArg a) {if (!l.blocked) (l.f1)(a);}
  };
  std::for_each(local_list.begin(), local_list.end(),
		std::bind2nd(std::ptr_fun(EmitItem::exec), arg));
}

template <class FreeArg>
FunctorArg<FreeArg> EmitterArg<FreeArg>::connect(const FunctorArg<FreeArg>& f1) {
  emission_list.push_back(ListItem(f1, Functor()));
  return f1;
}

template <class FreeArg>
FunctorArg<FreeArg> EmitterArg<FreeArg>::connect(const FunctorArg<FreeArg>& f1, Releaser& r) {
  // In this method:
  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed before the
  // remote object is
  // f3 is the functor the remote object executes in its Releaser if it is destroyed
  // before we are

  SafeFunctor f3(make(*this, &EmitterArg<FreeArg>::tracking_disconnect, f1));
  Functor f2(make(r, &Releaser::remove, f3));
  r.add(f3);
  try {
    emission_list.push_back(ListItem(f1, f2));
  }
  catch (...) {
    r.remove(f3);
    throw;
  }
  return f1;
}

template <class FreeArg>
void EmitterArg<FreeArg>::disconnect(const FunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, FunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove ourselves from the remote Releaser object
      (iter->f2)();

      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

// tracking disconnect() is the same as disconnect(), except that we do not
// execute f2 as the remote Releaser object will destroy its own functors
// in that case
template <class FreeArg>
void EmitterArg<FreeArg>::tracking_disconnect(FunctorArg<FreeArg> arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, FunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

template <class FreeArg>
void EmitterArg<FreeArg>::block(const FunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, FunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = true;
      ++iter;
    }
    else break;
  }
}

template <class FreeArg>
void EmitterArg<FreeArg>::unblock(const FunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, FunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = false;
      ++iter;
    }
    else break;
  }
}

template <>
class EmitterArg<void> {

  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed
  // before the remote object is
  struct ListItem {
    Functor f1;
    Functor f2;
    bool blocked;
    ListItem(Functor f1_, Functor f2_): f1(f1_), f2(f2_), blocked(false) {}
  };
  
  std::list<ListItem> emission_list;

  // only Releaser objects can access this
  void tracking_disconnect(Functor);

  // emitters cannot be copied
  EmitterArg(const EmitterArg&);
  EmitterArg& operator=(const EmitterArg&);
public:
  friend class Releaser;

  void operator()() const {emit();}
  void emit() const;

  // connect() returns the connected functor for convenience
  Functor connect(const Functor&);
  Functor connect(const Functor&, Releaser&);
  void disconnect(const Functor&);

  void block(const Functor&);
  void unblock(const Functor&);

  EmitterArg() {}
  ~EmitterArg();
};

inline EmitterArg<void>::~EmitterArg() {

  struct DisconnectReleaserItem {
    static void exec(const ListItem& l) {(l.f2)();}
  };
  std::for_each(emission_list.begin(), emission_list.end(),
		DisconnectReleaserItem::exec);
}

inline void EmitterArg<void>::emit() const {

  // create a local copy of emission_list, to enable a connected
  // function (i) to delete the EmitterArg<> object to which it is
  // connected, even if there are other functors still to execute in
  // the same emission (which will execute normally provided they do
  // not try to call any of the emitter's functions), (ii) to call
  // 'delete this' nothwithstanding that the connected function is
  // protected by a Releaser object (assuming all the other restraints
  // on calling 'delete this' are met), provided that no other access
  // would be made to the deleted object in a function call connected
  // to the same emitter which is due to execute subsequently in the
  // same emission, and (iii) to disconnect itself from the
  // EmitterArg<> object.  This design approach has a trade-off: if a
  // connected function tries to block, unblock or disconnect another
  // function connected to the same EmitterArg<> object which is due
  // to execute subsequently in the same emission (or to block,
  // unblock or disconnect itself when it is due to execute again
  // subsequently in the same emission), the attempted block, unblock
  // or disconnection will not have any effect on that emission (it
  // will only have effect on a subsequent emission).  In addition, a
  // connected function may not destroy an object whose non-static
  // method is connected to the same emitter and which would execute
  // subsequently in the same emission, even if that object is
  // protected by a Releaser object (the non-static method will
  // unsuccessfully attempt to execute notwithstanding the destruction
  // of the object it would be operating on).

  std::list<ListItem> local_list = emission_list;
  // we don't have lambdas in C++03, so use a struct in function scope
  struct EmitItem {
    static void exec(const ListItem& l) {if (!l.blocked) (l.f1)();}
  };
  std::for_each(local_list.begin(), local_list.end(),
		EmitItem::exec);
}

inline Functor EmitterArg<void>::connect(const Functor& f1) {
  emission_list.push_back(ListItem(f1, Functor()));
  return f1;
}

inline Functor EmitterArg<void>::connect(const Functor& f1, Releaser& r) {
  // In this method:
  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed before the
  // remote object is
  // f3 is the functor the remote object executes in its Releaser if it is destroyed
  // before we are

  SafeFunctor f3(make(*this, &EmitterArg<void>::tracking_disconnect, f1));
  Functor f2(make(r, &Releaser::remove, f3));
  r.add(f3);
  try {
    emission_list.push_back(ListItem(f1, f2));
  }
  catch (...) {
    r.remove(f3);
    throw;
  }
  return f1;
}

inline void EmitterArg<void>::disconnect(const Functor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, Functor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
 std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove ourselves from the remote Releaser object
      (iter->f2)();

      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

// tracking disconnect() is the same as disconnect(), except that we do not
// execute f2 as the remote Releaser object will destroy its own functors
// in that case
inline void EmitterArg<void>::tracking_disconnect(Functor arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, Functor f) {
      return (p.f1 == f);
    }
  };

 std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

inline void EmitterArg<void>::block(const Functor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, Functor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = true;
      ++iter;
    }
    else break;
  }
}

inline void EmitterArg<void>::unblock(const Functor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, Functor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = false;
      ++iter;
    }
    else break;
  }
}

template <class FreeArg>
class SafeEmitterArg {

  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed
  // before the remote object is
  struct ListItem {
    SafeFunctorArg<FreeArg> f1;
    SafeFunctorArg<int*> f2;
    bool blocked;
    ListItem(SafeFunctorArg<FreeArg> f1_, SafeFunctorArg<int*> f2_): f1(f1_), f2(f2_), blocked(false) {}
  };
  
  std::list<ListItem> emission_list;
  mutable Thread::Mutex mutex; // make this mutable so we can lock/unlock in const methods

  // only Releaser objects can access this
  void tracking_disconnect(SafeFunctorArg<FreeArg>);

  // emitters cannot be copied
  SafeEmitterArg(const SafeEmitterArg&);
  SafeEmitterArg& operator=(const SafeEmitterArg&);
public:
  friend class Releaser;

  void operator()(FreeArg arg) const {emit(arg);}
  void emit(FreeArg arg) const;

  // connect() returns the connected functor for convenience
  SafeFunctorArg<FreeArg> connect(const SafeFunctorArg<FreeArg>&);
  SafeFunctorArg<FreeArg> connect(const SafeFunctorArg<FreeArg>&, Releaser&);
  void disconnect(const SafeFunctorArg<FreeArg>&);

  void block(const SafeFunctorArg<FreeArg>&);
  void unblock(const SafeFunctorArg<FreeArg>&);

  SafeEmitterArg() {}
  ~SafeEmitterArg();
};

template <class FreeArg>
SafeEmitterArg<FreeArg>::~SafeEmitterArg() {

  // go through emission_list() item by item, popping off the front and erasing
  // as we go in case Releaser::try_remove() fails to acquire the lock on one
  // of the iterations
  Thread::Mutex::Lock lock(mutex);
  while (!emission_list.empty()) {
    typename std::list<ListItem>::iterator iter = emission_list.begin();
    int result = 0; // f2 might be a no-op
    // remove ourselves from the remote Releaser object
    (iter->f2)(&result);
    if (!result) { // we got the Releaser mutex lock or no-op
      // now remove this item from emission_list
      emission_list.erase(iter);
    }
    else {
      mutex.unlock();
      // spin nicely
#ifdef UTILS_EMITTER_DONT_USE_SCHED_YIELD
      usleep(10);
#else
      sched_yield();
#endif
      mutex.lock();
    }
  }
}

template <class FreeArg>
void SafeEmitterArg<FreeArg>::emit(FreeArg arg) const {

  // create a local copy of emission_list, to enable a connected
  // function (i) to delete the EmitterArg<> object to which it is
  // connected, even if there are other functors still to execute in
  // the same emission (which will execute normally provided they do
  // not try to call any of the emitter's functions), (ii) to call
  // 'delete this' nothwithstanding that the connected function is
  // protected by a Releaser object (assuming all the other restraints
  // on calling 'delete this' are met), provided that no other access
  // would be made to the deleted object in a function call connected
  // to the same emitter which is due to execute subsequently in the
  // same emission, and (iii) to disconnect itself from the
  // EmitterArg<> object.  This design approach has a trade-off: if a
  // connected function tries to block, unblock or disconnect another
  // function connected to the same EmitterArg<> object which is due
  // to execute subsequently in the same emission (or to block,
  // unblock or disconnect itself when it is due to execute again
  // subsequently in the same emission), the attempted block, unblock
  // or disconnection will not have any effect on that emission (it
  // will only have effect on a subsequent emission).  In addition, a
  // connected function may not destroy an object whose non-static
  // method is connected to the same emitter and which would execute
  // subsequently in the same emission, even if that object is
  // protected by a Releaser object (the non-static method will
  // unsuccessfully attempt to execute notwithstanding the destruction
  // of the object it would be operating on).

  // SafeFunctorArg<> usage has the additional point that while an
  // emission is in course, another thread should not try to do any of
  // those things, or the same outcome will result.  Another thread
  // should leave alone objects connected to a SafeEmitterArg<> object
  // from the time of operator()() or emit() beginning to the time of
  // it ending, and not try to interfere.

  // a side effect of having a local list is that, as required, we
  // will not be holding our mutex when executing the functors it
  // contains.  It is OK having the functors in two different lists
  // which are potentially (when our mutex is released) in two
  // different threads, because the functors hold their
  // Callback::Callback objects by SharedLockPtr so their reference
  // count is protected (they are SafeFunctorArg<> functors).

  std::list<ListItem> local_list;
  { // scope block for mutex lock
    Thread::Mutex::Lock lock(mutex);
    local_list = emission_list;
  }

  // we don't have lambdas in C++03, so use a struct in function scope
  struct EmitItem {
    static void exec(ListItem l, FreeArg a) {if (!l.blocked) (l.f1)(a);}
  };
  std::for_each(local_list.begin(), local_list.end(),
		std::bind2nd(std::ptr_fun(EmitItem::exec), arg));
}

template <class FreeArg>
SafeFunctorArg<FreeArg> SafeEmitterArg<FreeArg>::connect(const SafeFunctorArg<FreeArg>& f1) {
  Thread::Mutex::Lock lock(mutex);
  emission_list.push_back(ListItem(f1, SafeFunctorArg<int*>()));
  return f1;
}

template <class FreeArg>
SafeFunctorArg<FreeArg> SafeEmitterArg<FreeArg>::connect(const SafeFunctorArg<FreeArg>& f1, Releaser& r) {
  // In this method:
  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed before the
  // remote object is
  // f3 is the functor the remote object executes in its Releaser if it is destroyed
  // before we are

  SafeFunctor f3(make(*this, &SafeEmitterArg<FreeArg>::tracking_disconnect, f1));
  SafeFunctorArg<int*> f2(make(r, &Releaser::try_remove, f3));
  // we can't call Releaser::add() when holding our mutex or we will
  // get out of order locking, as Releaser's mutex is held when it
  // calls SafeEmitterArg<FreeArg>::tracking_disconnect() via f3, and
  // we don't need to do so
  r.add(f3);
  Thread::Mutex::Lock lock(mutex);
  try {
    emission_list.push_back(ListItem(f1, f2));
  }
  catch (...) {
    mutex.unlock();
    r.remove(f3);
    mutex.lock();
    throw;
  }
  return f1;
}

template <class FreeArg>
void SafeEmitterArg<FreeArg>::disconnect(const SafeFunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object more than
  // once, so cater for that as well as Releaser::try_remove() failing
  Thread::Mutex::Lock lock(mutex);
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for(;;) {
    iter = std::find_if(iter, emission_list.end(),
                        std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      int result = 0; // f2 might be a no-op
      // remove ourselves from the remote Releaser object
      (iter->f2)(&result);
      if (!result) { // we got the Releaser mutex lock or no-op
	// now remove this item from emission_list
        iter = emission_list.erase(iter);
      }
      else {
        mutex.unlock();
        // spin nicely
#ifdef UTILS_EMITTER_DONT_USE_SCHED_YIELD
        usleep(10);
#else
        sched_yield();
#endif
	mutex.lock();
	// start again at the beginning - we have released the mutex
	// so our iterator may have become invalid
	iter = emission_list.begin();
      }
    }
    else break;
  }
}

// tracking disconnect() is the same as disconnect(), except that we do not
// execute f2 as the remote Releaser object will destroy its own functors
// in that case
template <class FreeArg>
void SafeEmitterArg<FreeArg>::tracking_disconnect(SafeFunctorArg<FreeArg> arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  Thread::Mutex::Lock lock(mutex);
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

template <class FreeArg>
void SafeEmitterArg<FreeArg>::block(const SafeFunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  Thread::Mutex::Lock lock(mutex);
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = true;
      ++iter;
    }
    else break;
  }
}

template <class FreeArg>
void SafeEmitterArg<FreeArg>::unblock(const SafeFunctorArg<FreeArg>& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctorArg<FreeArg> f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  Thread::Mutex::Lock lock(mutex);
  typename std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = false;
      ++iter;
    }
    else break;
  }
}

template <>
class SafeEmitterArg<void> {

  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed
  // before the remote object is
  struct ListItem {
    SafeFunctor f1;
    SafeFunctorArg<int*> f2;
    bool blocked;
    ListItem(SafeFunctor f1_, SafeFunctorArg<int*> f2_): f1(f1_), f2(f2_), blocked(false) {}
  };
  
  std::list<ListItem> emission_list;
  mutable Thread::Mutex mutex; // make this mutable so we can lock/unlock in const methods

  // only Releaser objects can access this
  void tracking_disconnect(SafeFunctor);

  // emitters cannot be copied
  SafeEmitterArg(const SafeEmitterArg&);
  SafeEmitterArg& operator=(const SafeEmitterArg&);
public:
  friend class Releaser;

  void operator()() const {emit();}
  void emit() const;

  // connect() returns the connected functor for convenience
  SafeFunctor connect(const SafeFunctor&);
  SafeFunctor connect(const SafeFunctor&, Releaser&);
  void disconnect(const SafeFunctor&);

  void block(const SafeFunctor&);
  void unblock(const SafeFunctor&);

  SafeEmitterArg() {}
  ~SafeEmitterArg();
};

inline SafeEmitterArg<void>::~SafeEmitterArg() {

  // go through emission_list() item by item, popping off the front and erasing
  // as we go in case Releaser::try_remove() fails to acquire the lock on one
  // of the iterations
  Thread::Mutex::Lock lock(mutex);
  while (!emission_list.empty()) {
    std::list<ListItem>::iterator iter = emission_list.begin();
    int result = 0; // f2 might be a no-op
    // remove ourselves from the remote Releaser object
    (iter->f2)(&result);
    if (!result) { // we got the Releaser mutex lock or no-op
      // now remove this item from emission_list
      emission_list.erase(iter);
    }
    else {
      mutex.unlock();
      // spin nicely
#ifdef UTILS_EMITTER_DONT_USE_SCHED_YIELD
      usleep(10);
#else
      sched_yield();
#endif
      mutex.lock();
    }
  }
}

inline void SafeEmitterArg<void>::emit() const {

  // create a local copy of emission_list, to enable a connected
  // function (i) to delete the EmitterArg<> object to which it is
  // connected, even if there are other functors still to execute in
  // the same emission (which will execute normally provided they do
  // not try to call any of the emitter's functions), (ii) to call
  // 'delete this' nothwithstanding that the connected function is
  // protected by a Releaser object (assuming all the other restraints
  // on calling 'delete this' are met), provided that no other access
  // would be made to the deleted object in a function call connected
  // to the same emitter which is due to execute subsequently in the
  // same emission, and (iii) to disconnect itself from the
  // EmitterArg<> object.  This design approach has a trade-off: if a
  // connected function tries to block, unblock or disconnect another
  // function connected to the same EmitterArg<> object which is due
  // to execute subsequently in the same emission (or to block,
  // unblock or disconnect itself when it is due to execute again
  // subsequently in the same emission), the attempted block, unblock
  // or disconnection will not have any effect on that emission (it
  // will only have effect on a subsequent emission).  In addition, a
  // connected function may not destroy an object whose non-static
  // method is connected to the same emitter and which would execute
  // subsequently in the same emission, even if that object is
  // protected by a Releaser object (the non-static method will
  // unsuccessfully attempt to execute notwithstanding the destruction
  // of the object it would be operating on).

  // SafeFunctorArg<> usage has the additional point that while an
  // emission is in course, another thread should not try to do any of
  // those things, or the same outcome will result.  Another thread
  // should leave alone objects connected to a SafeEmitterArg<> object
  // from the time of operator()() or emit() beginning to the time of
  // it ending, and not try to interfere.

  // a side effect of having a local list is that, as required, we
  // will not be holding our mutex when executing the functors it
  // contains.  It is OK having the functors in two different lists
  // which are potentially (when our mutex is released) in two
  // different threads, because the functors hold their
  // Callback::Callback objects by SharedLockPtr so their reference
  // count is protected (they are SafeFunctorArg<> functors).

  std::list<ListItem> local_list;
  { // scope block for mutex lock
    Thread::Mutex::Lock lock(mutex);
    local_list = emission_list;
  }

  // we don't have lambdas in C++03, so use a struct in function scope
  struct EmitItem {
    static void exec(const ListItem& l) {if (!l.blocked) (l.f1)();}
  };
  std::for_each(local_list.begin(), local_list.end(),
		EmitItem::exec);
}

inline SafeFunctor SafeEmitterArg<void>::connect(const SafeFunctor& f1) {
  Thread::Mutex::Lock lock(mutex);
  emission_list.push_back(ListItem(f1, SafeFunctorArg<int*>()));
  return f1;
}

inline SafeFunctor SafeEmitterArg<void>::connect(const SafeFunctor& f1, Releaser& r) {
  // In this method:
  // f1 is the functor we execute when we emit()
  // f2 is the functor we execute in our destructor if we are destroyed before the
  // remote object is, or in SafeEmitterArg<void>::disconnect()
  // f3 is the functor the remote object executes in its Releaser if it is destroyed
  // before we are

  SafeFunctor f3(make(*this, &SafeEmitterArg<void>::tracking_disconnect, f1));
  SafeFunctorArg<int*> f2(make(r, &Releaser::try_remove, f3));
  // we can't call Releaser::add() when holding our mutex or we will
  // get out of order locking, as Releaser's mutex is held when it
  // calls SafeEmitterArg<FreeArg>::tracking_disconnect() via f3, and
  // we don't need to do so
  r.add(f3);
  Thread::Mutex::Lock lock(mutex);
  try {
    emission_list.push_back(ListItem(f1, f2));
  }
  catch (...) {
    mutex.unlock();
    r.remove(f3);
    mutex.lock();
    throw;
  }
  return f1;
}

inline void SafeEmitterArg<void>::disconnect(const SafeFunctor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object more than
  // once, so cater for that as well as Releaser::try_remove() failing
  Thread::Mutex::Lock lock(mutex);
  std::list<ListItem>::iterator iter = emission_list.begin();
  for(;;) {
    iter = std::find_if(iter, emission_list.end(),
                        std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      int result = 0; // f2 might be a no-op
      // remove ourselves from the remote Releaser object
      (iter->f2)(&result);
      if (!result) { // we got the Releaser mutex lock or no-op
	// now remove this item from emission_list
        iter = emission_list.erase(iter);
      }
      else {
        mutex.unlock();
        // spin nicely
#ifdef UTILS_EMITTER_DONT_USE_SCHED_YIELD
        usleep(10);
#else
        sched_yield();
#endif
	mutex.lock();
	// start again at the beginning - we have released the mutex
	// so our iterator may have become invalid
	iter = emission_list.begin();
      }
    }
    else break;
  }
}

// tracking disconnect() is the same as disconnect(), except that we do not
// execute f2 as the remote Releaser object will destroy its own functors
// in that case
inline void SafeEmitterArg<void>::tracking_disconnect(SafeFunctor arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctor f) {
      return (p.f1 == f);
    }
  };

  Thread::Mutex::Lock lock(mutex);
  std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      // remove this item from emission_list
      iter = emission_list.erase(iter);
    }
    else break;
  }
}

inline void SafeEmitterArg<void>::block(const SafeFunctor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  Thread::Mutex::Lock lock(mutex);
  std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = true;
      ++iter;
    }
    else break;
  }
}

inline void SafeEmitterArg<void>::unblock(const SafeFunctor& arg) {
  // we don't have lambdas in C++03, so use a struct in function scope
  struct Pred {
    // we can't pass const reference types as we bind with std::bind2nd below
    static bool pred(ListItem p, SafeFunctor f) {
      return (p.f1 == f);
    }
  };

  // in theory, we could have connected the same functor object
  // more than once, so cater for that
  Thread::Mutex::Lock lock(mutex);
  std::list<ListItem>::iterator iter = emission_list.begin();
  for (;;) {
    iter = std::find_if(iter, emission_list.end(),
			std::bind2nd(std::ptr_fun(Pred::pred), arg));
    if (iter != emission_list.end()) {
      iter->blocked = false;
      ++iter;
    }
    else break;
  }
}

} // namespace Callback

#endif
